"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
exports.__esModule = true;
var chalk_1 = require("chalk");
var cli = require("commander");
var _ = require("lodash");
var coercion_1 = require("./coercion");
var errors_1 = require("./errors");
var metadata_1 = require("./metadata");
var iocContainer;
var commandDefinitions = [];
function setIocContainer(container) {
    iocContainer = container;
}
exports.setIocContainer = setIocContainer;
function getIocContainer() {
    return iocContainer;
}
exports.getIocContainer = getIocContainer;
function getOptionUsage(option) {
    var result = '';
    if (option.shortName) {
        result = "-" + option.shortName + ", ";
    }
    result += "--" + option.name.toString() + " ";
    if (option.valueName) {
        result += "<" + option.valueName + ">";
    }
    return result.trim();
}
function getParams(command, args) {
    var values = metadata_1.getCommandValues(command.paramsClass.prototype);
    var options = metadata_1.getCommandOptions(command.paramsClass.prototype);
    var params = new command.paramsClass();
    var paramIndex = 0;
    for (var _i = 0, values_1 = values; _i < values_1.length; _i++) {
        var value = values_1[_i];
        var argValue = args[paramIndex++];
        var defaultValue = params[value.name];
        var paramValue = coercion_1.coerceValue(argValue, value.type, defaultValue);
        params[value.name] = paramValue;
    }
    var optionValues = args[paramIndex];
    for (var _a = 0, options_1 = options; _a < options_1.length; _a++) {
        var option = options_1[_a];
        var value = optionValues[option.name];
        if (value !== undefined) {
            params[option.name.toString()] = value;
        }
    }
    return params;
}
function instantiateCommand(command) {
    var constructor = command.type;
    var instance = iocContainer ? iocContainer.get(constructor) : new constructor();
    return instance;
}
function registerCommandOption(cliCommand, paramsClass, option) {
    var optionUsage = getOptionUsage(option);
    var coercedValue = !option.valueName ? undefined : (function (value) {
        if (option.valueName && option.type === Number) {
            return +value;
        }
        else {
            return value;
        }
    });
    var params = new paramsClass();
    var defaultValue = params[option.name];
    cliCommand.option(optionUsage, option.description, coercedValue, defaultValue);
}
function registerCommand(commandDefinition) {
    var _this = this;
    commandDefinitions.push(commandDefinition);
    var usage = getCommandDefinitionUsage(commandDefinition);
    var cliCommand = cli.command(usage);
    if (commandDefinition.description) {
        cliCommand.description(commandDefinition.description);
    }
    var options = metadata_1.getCommandOptions(commandDefinition.paramsClass.prototype);
    for (var _i = 0, options_2 = options; _i < options_2.length; _i++) {
        var option = options_2[_i];
        registerCommandOption(cliCommand, commandDefinition.paramsClass, option);
    }
    cliCommand.action(function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return __awaiter(_this, void 0, void 0, function () {
            var commandInstance, params, err_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 2, , 3]);
                        commandInstance = instantiateCommand(commandDefinition);
                        params = getParams(commandDefinition, args);
                        return [4 /*yield*/, commandInstance.execute(params)];
                    case 1:
                        _a.sent();
                        return [3 /*break*/, 3];
                    case 2:
                        err_1 = _a.sent();
                        // tslint:disable:no-console
                        console.error();
                        console.error(chalk_1["default"].red(errors_1.errorToString(err_1)));
                        console.error();
                        process.exit(1);
                        return [3 /*break*/, 3];
                    case 3: return [2 /*return*/];
                }
            });
        });
    });
}
exports.registerCommand = registerCommand;
function valueDefinitionToUsageString(value) {
    var name = "" + value.name + (value.variadic ? '...' : '');
    var usage = value.optional ? "[" + name + "]" : "<" + name + ">";
    return usage;
}
function getCommandDefinitionUsage(commandDefinition) {
    var values = metadata_1.getCommandValues(commandDefinition.paramsClass.prototype);
    var valuesUsage = _
        .chain(values)
        .map(function (value, index) { return ({
        value: value,
        index: index
    }); })
        .sort(function (a, b) { return a.value.variadic ? 1 : b.value.variadic ? -1 : a.index - b.index; })
        .map(function (valueAndIndex) { return valueDefinitionToUsageString(valueAndIndex.value); })
        .join(' ')
        .value();
    var usage = (commandDefinition.name + " " + valuesUsage).trim();
    return usage;
}
exports.getCommandDefinitionUsage = getCommandDefinitionUsage;
function getCommandUsage(commandClass) {
    var commandDefinition = commandDefinitions
        .find(function (item) { return item.type === commandClass; });
    if (!commandDefinition) {
        throw new Error('Class is not a command');
    }
    var usage = getCommandDefinitionUsage(commandDefinition);
    return usage;
}
exports.getCommandUsage = getCommandUsage;
